# -*- coding: utf-8 -*-

import asyncio
import json
import os
import io
from urllib.parse import urlparse

from .exceptions import RateLimitExceeded

try:
    from magic import Magic
    mime = Magic(mime=True)
    magic = True
except:
    print('Could not load python-magic, fallback to mimetypes')
    import mimetypes
    mime = mimetypes.MimeTypes()
    magic = False

from PIL import Image


class JSONObject(dict):
    """
        A dict in which you can access items as attributes

    >>> obj = JSONObject(key=True)
    >>> obj['key'] is obj.key  # returns True
    """

    def __getattr__(self, key):
        if key in self:
            return self[key]
        raise AttributeError("%s has no property named %s." %
                             (self.__class__.__name__, key))

    def __setattr__(self, *args):
        raise AttributeError("%s instances are read-only." %
                             self.__class__.__name__)
    __delattr__ = __setitem__ = __delitem__ = __setattr__


class PeonyResponse:
    """
        Response objects

    In these object you can access the headers, the request, the url
    and the response
    getting an attribute/item of this object will get the corresponding
    attribute/item of the response

    >>> peonyresponse.key is peonyresponse.response.key  # returns True
    >>>
    >>> # iterate over peonyresponse.response
    >>> for key in peonyresponse:
    ...     pass  # do whatever you want
    """

    def __init__(self, response, headers, url, request):
        """ keep informations about the response as instance attributes """
        self.response = response
        self.headers = headers
        self.url = url
        self.request = request

    def __getattr__(self, key):
        return getattr(self.response, key)

    def __getitem__(self, key):
        return self.response[key]

    def __iter__(self):
        return iter(self.response)

    def __str__(self):
        return str(self.response)

    def __repr__(self):
        return repr(self.response)

    def __len__(self):
        return len(self.response)


def handler_decorator(handler):
    def decorated_handler(request):
        def decorated_request(error_handling=True, **kwargs):
            if error_handling:
                return handler(request)(**kwargs)
            else:
                return request(**kwargs)

        return decorated_request

    return decorated_handler


@handler_decorator
def error_handler(request):
    async def decorated_request(error_handling=True, **kwargs):
        while True:
            try:
                return await request(**kwargs)
            except RateLimitExceeded as e:
                print(e)
                delay = int(e.reset_in) + 1
                print("sleeping for %ds" % delay)
                await asyncio.sleep(delay)
            except TimeoutError:
                print("connection timed out")
            else:
                raise

    return decorated_request


def loads(json_data, *args, encoding="utf-8", **kwargs):
    """ custom loads function with an object_hook and automatic decoding """
    if isinstance(json_data, bytes):
        json_data = json_data.decode(encoding)

    return json.loads(json_data, *args, object_hook=JSONObject, **kwargs)


def media_chunks(media, chunk_size, media_size):
    while media.tell() < media_size:
        yield media.read(chunk_size)


def convert(img, formats):
    for kwargs in formats:
        f = io.BytesIO()
        img.save(f, **kwargs)
        yield f


def optimize_media(path, max_size, formats):
    with Image.open(path) as img:
        ratio = max(hw / max_hw for hw, max_hw in zip(img.size, max_size))

        if ratio > 1:
            size = tuple(int(hw // ratio) for hw in img.size)
            img = img.resize(size, Image.ANTIALIAS)

        files = list(convert(img, formats))

    files.sort(key=get_size)
    media = files.pop(0)

    for f in files:
        f.close()

    return media


def reset_io(func):
    def decorated(media):
        media.seek(0)
        result = func(media)
        media.seek(0)

        return result

    return decorated


@reset_io
def get_media_metadata(f):
    media_type, media_category = get_type(f)
    is_image = not (media_type.endswith('gif')
                    or media_type.startswith('video'))

    return media_type, media_category, is_image


def get_image_metadata(f):
    # try to get the path no matter how the input is
    if isinstance(f, str):
        path = urlparse(f).path.strip(" \"'")

        with open(path, 'rb') as original:
            return (*get_media_metadata(original), path)

    elif hasattr(f, 'read'):
        if hasattr(f, 'name'):
            path = urlparse(f.name).path.strip(" \"'")
        else:
            path = None

        return (*get_media_metadata(f), path)
    else:
        raise TypeError("upload_media input must be a file object or a"
                        "filename")


@reset_io
def get_size(media):
    media.seek(0, os.SEEK_END)
    return media.tell()


@reset_io
def get_type(media, path=None):
    if magic:
        media_type = mime.from_buffer(media.read(1024))
    elif path:
        media_type = mime.guess_type(path)
    else:
        raise RuntimeError("Cannot guess mimetype of media")

    if media_type.startswith('video'):
        media_category = "tweet_video"
    elif media_type.endswith('gif'):
        media_category = "tweet_gif"
    else:
        media_category = "tweet_image"

    return media_type, media_category



# -*- coding: utf-8 -*-
"""  peony-twitter

    An asynchronous Twitter API client for Python
"""

__author__ = "Florian Badie"
__author_email__ = "florianbadie@gmail.com"
__url__ = "https://github.com/odrling/peony-twitter"

__version__ = "0.3"

__license__ = "MIT License"

__keywords__ = "twitter, asyncio, asynchronous"

from .client import PeonyClient, BasePeonyClient
from .commands import EventStream, task, event_handler, events
from .utils import handler_decorator



# -*- coding: utf-8 -*-

import re
import sys

from . import utils


class Functions(dict):

    def __init__(self, *args, prefix=None, strict=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.prefix = prefix
        ex = re.escape(prefix) + "[^ \n\t]+"
        self.prog = re.compile(ex)
        self.strict = strict

    def __getitem__(self, k):
        if not isinstance(k, str):
            msg = "%s key must be a string" % self.__class__.__name__
            raise ValueError(msg)

        if not k.startswith(self.prefix):
            k = self.prefix + k

        return super().__getitem__(k)

    def __setitem__(self, k, v):
        if self.prefix is None:
            raise RuntimeError("To add commands you must set a prefix")

        if not isinstance(k, str):
            msg = "%s key must be a string" % self.__class__.__name__
            raise ValueError(msg)

        if not k.startswith(self.prefix):
            k = self.prefix + k

        super().__setitem__(k, v)

    def __repr__(self):
        return "{classname}({super})".format(
            classname=self.__class__.__name__,
            super=", ".join(super().keys())
        )

    def _get(self, text):
        if self.strict:
            match = self.prog.match(text)
            if match:
                cmd = match.group()
                if cmd in self:
                    return cmd
        else:
            words = self.prog.findall(text)
            for word in words:
                if word in self:
                    return word

    async def run(self, *args, data, **kwargs):
        cmd = self._get(data.text)

        try:
            if cmd is not None:
                command = self[cmd](*args, data=data, **kwargs)
                return await utils.execute(command)

        except Exception as e:
            fmt = "Error occurred while running function {cmd}: {error}"
            print(fmt.format(cmd=cmd, error=e), file=sys.stderr)

    def __call__(self, func, name=None):
        name = name or func.__name__
        self[name] = func

        return func


class Commands(Functions):

    def __init__(self, prefix=None):
        super().__init__(prefix=prefix)

        @self
        def help(_self, data, *args, **kwargs):
            """ show commands help """
            kdoc = [
                (key, utils.doc(value))
                for key, value in self.items()
                if utils.permission_check(data,
                                          _permissions=_self.permissions,
                                          command=value)
            ]

            kdoc.sort(key=self._key)

            msg = ["`{key}`: {doc}".format(key=key, doc=doc)
                   for key, doc in kdoc]

            return "\n".join(msg)

    def _key(self, item):
        key, __ = item
        if key == '%shelp' % self.prefix:
            return ''   # /help is the first in the message
        else:
            return key  # sort other commands alphabeticaly

    def restricted(self, *permissions):
        def decorator(func):
            async def decorated(_self, data, *args, **kwargs):
                permission = utils.permission_check(
                    data,
                    _permissions=_self.permissions,
                    permissions=permissions
                )

                if permission:
                    cmd = func(_self, data, *args, **kwargs)

                    return await utils.execute(cmd)

            decorated.__doc__ = func.__doc__
            decorated.permissions = permissions

            return self(decorated, name=func.__name__)

        return decorator



# -*- coding: utf-8 -*-
import asyncio

from . import utils


class BaseIterator:

    def __init__(self, _request, **kwargs):
        self.request = _request
        self.kwargs = kwargs

    def __aiter__(self):
        return self

    async def __anext__(self):
        pass


class IdIterator(BaseIterator):
    """
        Iterate using ids

    It is the parent class of MaxIdIterator and SinceIdIterator
    """

    def __init__(self, _request, _parameter, _i, _force=False, **kwargs):
        """ Keep all the arguments as class attributes """
        self.param = _parameter
        self.force = _force
        self.i = _i
        super().__init__(_request, **kwargs)

    async def __anext__(self):
        """ return each response until getting an empty response """
        response = await self.request(**self.kwargs)

        if response:
            response = await self.call_on_response(response)
        elif not self.force:
            raise StopAsyncIteration

        return response

    async def call_on_response(self, response):
        """
            The parameter is set to the id of the tweet at index i - 1
        """
        self.kwargs[self.param] = response[self.i]['id'] - 1
        return response


class MaxIdIterator(IdIterator):
    """
        Iterator for endpoints using max_id
    """

    def __init__(self, _request, **kwargs):
        """ Keep all the arguments as class attributes """
        super().__init__(_request,
                         _parameter="max_id",
                         _i=-1,
                         _force=False,
                         **kwargs)


class SinceIdIterator(IdIterator):
    """
        Iterator for endpoints using since_id
    """

    def __init__(self, _request, _force=True, _fill_gaps=True, **kwargs):
        super().__init__(_request,
                         _parameter="since_id",
                         _i=0,
                         _force=_force,
                         **kwargs)

        self.fill_gaps = _fill_gaps

    async def set_param(self, response):
        if self.fill_gaps:
            self.kwargs[self.param] = response[self.i]['id'] - 1
        else:
            self.kwargs[self.param] = response[self.i]['id']

    async def call_on_response(self, response):
        """
            Try to fill the gaps and strip last tweet from the response
            if its id is that of the first tweet of the former response
        """
        since_id = self.kwargs.get(self.param, 0)
        if self.fill_gaps:
            since_id += 1

        if since_id != response[-1]['id']:
            if self.fill_gaps:
                responses = with_max_id(
                    _request=self.request,
                    max_id=response[-1]['id'] - 1,
                    **self.kwargs
                )

                async for tweets in responses:
                    response.extend(tweets)

            if since_id != response[-1]['id']:
                await self.set_param(response)
                return response

        await self.set_param(response)
        return response[:-1]


class CursorIterator(BaseIterator):
    """ Iterate using a cursor """

    async def __anext__(self):
        """ return each response until getting 0 as next cursor """
        if self.kwargs.get('cursor', -1) != 0:
            response = await self.request(**self.kwargs)
            self.kwargs['cursor'] = response['next_cursor']
            return response
        else:
            raise StopAsyncIteration


with_max_id = MaxIdIterator
with_since_id = SinceIdIterator
with_cursor = CursorIterator



# -*- coding: utf-8 -*-
import time

from . import utils


def get_error(data):
    """ Get the error, this is quite a vertical code """
    if data is not None:
        if 'errors' in data:
            if data['errors']:
                return data['errors'][0]


async def throw(response, **kwargs):
    """ get the response data if possible and raise an exception """
    ctype = response.headers['CONTENT-TYPE'].lower()
    data = None

    if "json" in ctype:
        try:
            data = await response.json(loads=utils.loads)
        except:
            pass

    err = get_error(data)
    if err is not None:
        if 'code' in err:
            code = str(err['code'])
            if code in error:
                e = error[code]
                raise e(response=response, data=data, **kwargs)

    if str(response.status) in status:
        e = status[response.status]
        raise e(response=response, data=data, **kwargs)

    # raise PeonyException if no specific exception was found
    raise PeonyException(response=response, data=data, **kwargs)


class PeonyException(Exception):
    """ Parent class of all the exceptions of Peony """

    def __init__(self, response=None, data=None, message=None):
        """
            Add the response and data attributes

        Extract message from the error if not explicitly given
        """
        if not message:
            err = get_error(data)
            if err is not None:
                if 'message' in err:
                    message = err['message']

            message = message or str(response)

        self.response = response
        self.data = data

        super().__init__(message)


class MediaProcessingError(PeonyException):
    pass


class StreamLimit(PeonyException):
    pass


def convert_int_keys(func):
    def decorated(self, key, *args, **kwargs):
        if isinstance(key, int):  # convert int keys to str
            key = str(key)

        return func(self, key, *args, **kwargs)

    return decorated


class ErrorDict(dict):
    """ A dict to easily add exception associated to a code """

    @convert_int_keys
    def __getitem__(self, key):
        return super().__getitem__(key)

    @convert_int_keys
    def __setitem__(self, key, value):
        super().__setitem__(key, value)

    def code(self, code):
        def decorator(exception):
            self[code] = exception
            return exception

        return decorator


status = ErrorDict()
error = ErrorDict()


@error.code(32)
class NotAuthenticated(PeonyException):
    pass


@error.code(34)
class DoesNotExist(PeonyException):
    pass


@error.code(64)
class AccountSuspended(PeonyException):
    pass


@error.code(68)
class MigrateToNewAPI(PeonyException):
    pass


@error.code(88)
class RateLimitExceeded(PeonyException):

    @property
    def reset(self):
        return int(self.response.headers['X-Rate-Limit-Reset'])

    @property
    def reset_in(self):
        return self.reset - time.time()


@error.code(92)
class SSLRequired(PeonyException):
    pass


@error.code(130)
class OverCapacity(PeonyException):
    pass


@error.code(131)
class InternalError(PeonyException):
    pass


@error.code(135)
class CouldNotAuthenticate(PeonyException):
    pass


@error.code(136)
class Blocked(PeonyException):
    pass


@error.code(161)
class FollowLimit(PeonyException):
    pass


@error.code(179)
class ProtectedTweet(PeonyException):
    pass


@error.code(185)
class StatusLimit(PeonyException):
    pass


@error.code(187)
class DuplicatedStatus(PeonyException):
    pass


@error.code(215)
class BadAuthentication(PeonyException):
    pass


@error.code(226)
class AutomatedRequest(PeonyException):
    pass


@error.code(231)
class VerifyLogin(PeonyException):
    pass


@error.code(251)
class RetiredEndpoint(PeonyException):
    pass


@error.code(261)
class ReadOnlyApplication(PeonyException):
    pass


@error.code(271)
class CannotMuteYourself(PeonyException):
    pass


@error.code(272)
class NotMutingUser(PeonyException):
    pass


@error.code(354)
class DMCharacterLimit(PeonyException):
    pass


@status.code(304)
class NotModified(PeonyException):
    pass


@status.code(400)
class BadRequest(PeonyException):
    pass


@status.code(401)
class Unauthorized(PeonyException):
    pass


@status.code(403)
class Forbidden(PeonyException):
    pass


@status.code(404)
class NotFound(PeonyException):
    pass


@status.code(406)
class NotAcceptable(PeonyException):
    pass


@status.code(410)
class Gone(PeonyException):
    pass


@status.code(420)
class EnhanceYourCalm(PeonyException):
    pass


@status.code(422)
class UnprocessableEntity(PeonyException):
    pass


@status.code(429)
class TooManyRequests(PeonyException):
    pass


@status.code(500)
class InternalServerError(PeonyException):
    pass


@status.code(502)
class BadGateway(PeonyException):
    pass


@status.code(503)
class ServiceUnavailable(PeonyException):
    pass


@status.code(504)
class GatewayTimeout(PeonyException):
    pass



# -*- coding: utf-8 -*-

import asyncio
import json
import sys

import aiohttp

from . import utils, exceptions
from .exceptions import StreamLimit, PeonyException
from .general import rate_limit_notices


class StreamResponse:
    """
        Asynchronous iterator for streams
    """

    def __init__(self, *args, _headers,
                 session=None,
                 reconnect=150,
                 loads=utils.loads,
                 timeout=90,
                 **kwargs):
        """ keep the arguments as instance attributes """
        self.session = session or aiohttp.ClientSession()
        self.reconnect = reconnect
        self.headers = _headers
        self.args = args
        self.kwargs = kwargs
        self.loads = loads
        self.timeout = timeout

    async def __aiter__(self):
        """ create the connection """
        kwargs = self.headers.prepare_request(**self.kwargs)

        self.response = await self.session.request(*self.args, **kwargs)
        if self.response.status == 200:
            return self
        else:
            try:
                await exceptions.throw(self.response)
            except PeonyException as e:
                print(e, file=sys.stderr)
                if self.reconnect:
                    await self.restart_stream(error=e)

    async def __anext__(self):
        """ decode each line using json """
        line = b''
        try:
            while not line:
                coro = self.response.content.__aiter__().__anext__()
                line = await asyncio.wait_for(coro, self.timeout)
                line = line.rstrip(b'\r\n')

            if line in rate_limit_notices:
                raise StreamLimit(line)

            return self.loads(line)

        except StreamLimit as error:
            print("Error:", line, file=sys.stderr)
            return await self.restart_stream(error=error)

        except StopAsyncIteration as error:
            print("Stream stopped", file=sys.stderr)
            return await self.restart_stream(error=error)

        except json.decoder.JSONDecodeError as error:
            print("Decode error:", line, file=sys.stderr)
            return await self.restart_stream(error=error)

        except asyncio.TimeoutError as error:
            print("Timeout reached", file=sys.stderr)
            return await self.restart_stream(reconnect=0, error=error)

        except GeneratorExit:
            self.response.close()
            self.session.close()

        except KeyboardInterrupt:
            self.response.close()
            self.session.close()

        except Exception as e:
            str(e) and print(e, file=sys.stderr)
            self.response.close()
            self.session.close()
            raise

    async def restart_stream(self, reconnect=None, error=None):
        """ restart the stream on error """

        if error is not None:
            print(error, file=sys.stderr)

        reconnect = reconnect is None and self.reconnect or reconnect

        try:
            self.response.close()
        except Exception as e:
            print(e, file=sys.stderr)

        if reconnect is not None:
            if reconnect > 0:
                print("restarting stream in %ss" % reconnect)
                await asyncio.sleep(reconnect)

            print("restarting stream")
            await self.__aiter__()
            return await self.__anext__()
        else:
            raise error


class StreamContext:
    """ A context that should close the session on exit """

    def __init__(self, *args, **kwargs):
        """ keep the arguments as instance attributes """
        self.args = args
        self.kwargs = kwargs

    async def __aenter__(self):
        """ create stream and return it """
        self.stream = StreamResponse(*self.args, **self.kwargs)

        return self.stream

    async def __aexit__(self, *args, **kwargs):
        """ close the response and the session """
        if hasattr(self.stream, "response"):
            self.stream.response.close()
        if hasattr(self.stream, "session"):
            self.stream.session.close()

        return True
